 /*
 * Ethnocentricity Model
 *
 * PLAYER code
 *
 * "Additive" variant (new REGULAR case), originally coded 3/12/02
 * As per Ethn29 memo, 3/7/02
 *
 * This code version used for paper submitted to SCIENCE, 4/02
 *
 * Additions 5/9/02:
 *	- combat variant and switch
 *  - cost to discrim and switch
 *  - cleaned up code comments for documentation purposes
 *
 */

package edu.brook.ethnoadd;

import java.awt.*;
import java.util.*;

import edu.brook.ascape.model.*;
import edu.brook.ascape.rule.*;

import edu.brook.ascape.util.*;
import edu.brook.ascape.view.*;

/**
 * A player in the ethnocentricity model.
 *
 **/

public class EthnoPlayer extends CellOccupant   {
	
	
	//constants
	protected final static int BLUE = 0;
	protected final static int RED = 1;
    protected final static int GREEN = 2;
    protected final static int YELLOW = 3;
    
    protected final int MEETOWN = 0;
	protected final int MEETOTHER = 1;
	protected final int DEFOTHER = 2;
	protected final int COOPOWN = 3;
    protected final int COOPOTHER = 4;
    
    //agent fields
    protected int tag;
	protected boolean strategyToOwn;  //TRUE = COOPERATE
	protected boolean strategyToOther;  //TRUE = COOPERATE
		
    protected double currOTR;

		
	//INITIALIZE AN AGENT (clone or immigrant)	
	public void initialize() {
		super.initialize();
   
        //check for OneTagOnly
        if ((((EthnoModel) scape.getModel()).getOneTagOnly())) {
            tag = BLUE;
        }
        else {
            if (((EthnoModel) scape.getModel()).getUseMoreColors()) {
         		int numColors = ((EthnoModel) scape.getModel()).getNumColors();
                tag = randomInRange(0,numColors);    //0,3 normal   //0,7 for 8 colors
            }
            else {
                if (randomIs()) {
                    tag = BLUE;
                }
                else {
                    tag = RED;
                }
            }
        }
        
        strategyToOwn = randomIs();
        strategyToOther = randomIs();
        //strategyToOther = strategyToOwn;  //use for an all CC or DD start
        

	}
    
    
    
	public void scapeCreated() {
		
		getScape().addInitialRule(MOVE_RANDOM_LOCATION_RULE);
	    getScape().addRule(UPDATE_RULE); //Resets OTR at beginning of round

        
        //The code below adds various "play" or interaction rules
        //Only one of the three rules below should be active (the
        //other two should be commented out)
                
        //USE THIS RULE FOR "REGULAR CASE" 
        getScape().addRule(PLAY_NEIGHBORS_RULE); //play all neighbors
        
        //USE THIS RULE FOR RANDOM LOCATION VARIANT        
        /*
        getScape().addRule(new Rule("Play a Random Partner") {       
            public void execute(Agent a) {
                 playRandomGlobal(a);
             }
        });
        */
        
        
        //USE THIS RULE FOR PLAY ONE NEIGHBOR ONLY VARIANT
        /*
        getScape().addRule(new Rule("Play a Random Neighbor") {
		    public void execute(Agent a) {
		        Cell[] neighbors = ((EthnoPlayer) a).getHostCell().getNeighbors();
		        
		        boolean neighborAvail = false;       
		        for (int i=0; i<neighbors.length; i++) {
		            if (!(neighbors[i].isAvailable())) {
		                neighborAvail = true;
		            }
		        }
		        if (neighborAvail) {
		            boolean done = false;
		            while (!(done)) {
		                int pick = randomToLimit(neighbors.length);
		                if (!(neighbors[pick].isAvailable())) {
		                    done = true;
		                    EthnoPlayer partner = ((EthnoPlayer) (neighbors[pick].getOccupant()));
		                    ((EthnoPlayer) a).play(partner);
		                }
		            }
		        }
		    }
		});
		*/
	
		getScape().addRule(FISSIONING_RULE); //This rule handles cloning

   		getScape().addRule(DEATH_RULE); //This rule handles death of agents
		
	}
    
    
    public void playRandomGlobal(Agent a) { //Active only if specified above
        
        EthnoPlayer candidatePartner = null;
		do {
		    //Find random candidate uniformally distributed across all agents
			int index = randomToLimit(scape.getSize());
			candidatePartner = (EthnoPlayer) ((ScapeVector) scape).getCell(index);
		} while ((candidatePartner == this));
		
		if (candidatePartner.getHostCell().isAvailable()) {
		    System.out.println("EMPTY PARTNER!!!");
		}
		
		
        a.play(candidatePartner);
        
    }
    
    
    public void update() { //RESET OTR TO BASE RATE
        this.currOTR = ((EthnoModel) scape.getModel()).getBaseOTR(); 
        
        if (((EthnoModel) scape.getModel()).getUseDiscrimCost()) { 
        	if (!(strategyToOwn == strategyToOther)) {
        		this.currOTR -= ((EthnoModel) scape.getModel()).getDiscrimCost();
        	}        	
        }
        
   
    }
   
   
    public boolean fissionCondition() {//FLIP BASED ON OTR FOR CLONING CHANCE
            
       if (randomInRange(0.0,1.0) < this.currOTR) {
            return (true);
       }
       else {
            return (false);
       }
        
    }
    
    
    public boolean deathCondition() { //FLIP BASED ON DEATH RATE FROM MAIN SETTINGS
        
        double theDeathRate = ((EthnoModel) scape.getModel()).getDeathRate();



        if (randomInRange(0.0,1.0) < theDeathRate) {
            return (true);
        }
        else {
            return (false);
        }
        
    }
    
        
        
    public void fission() { //CLONING ROUTINE (active if fired by FissionCondition)
                
       
                
        boolean locateChildNear = ((EthnoModel) scape.getModel()).getLocateChildNear();
    
        //LOCAL REPRODUCTION -- CLONES NEAR PARENT
        if (locateChildNear) {
            if (this.getHostCell().isNeighborAvailable()) {
    
    		   
                //create offspring
                EthnoPlayer child = (EthnoPlayer) this.clone();
                ((ScapeVector) getScape()).addAgent(child);
                child.moveTo(getHostCell().findRandomAvailableNeighbor());
      
                //mutate strategyToOwn?
                if (getRandom().nextFloat() < ((EthnoModel) scape.getModel()).getMutationRate()) {
                    child.strategyToOwn = !(this.strategyToOwn);
                }
                else {
                    child.strategyToOwn = this.strategyToOwn;
                }
                
                //mutate strategyToOther?
                if (getRandom().nextFloat() < ((EthnoModel) scape.getModel()).getMutationRate()) {
                    child.strategyToOther = !(this.strategyToOther);    
                }
                else {
                    child.strategyToOther = this.strategyToOther;
                }
                
                //mutate tag?
                if ((((EthnoModel) scape.getModel()).getOneTagOnly())) { //ONE COLOR CASE
                    child.tag = this.tag;
                }
                else {//MULTIPLE COLORS
                    if (getRandom().nextFloat() < ((EthnoModel) scape.getModel()).getMutationRate()) {    
                                
                        if (((EthnoModel) scape.getModel()).getUseMoreColors()) {
                
                			int numColors = ((EthnoModel) scape.getModel()).getNumColors();
                		    child.tag = randomInRange(0,numColors); //0,3 normal //0,7 for 8 colors

                            while (child.tag == this.tag) {
                                child.tag = randomInRange(0,numColors); //0,3 normal  //0,7 for 8 colors
                            }
                            
                        }
                        else { //2 colors
                            if (this.tag == BLUE) {child.tag = RED;}                    
                            else {child.tag = BLUE;}
                                       
                        }
                    }
                    else {child.tag = this.tag;} //Don't mutate
                }
                
                //This next line is for Cell statistics... not currently
                //active, but left in case might be useful later
                ((EthnoCell) child.getHostCell()).updateCell();
   
            }//end if neighbor avail
        }//end if(locateChildNear)
        else { //RANDOM CLONING - Locate child in random location if avail.
            
            if (((EthnoModel) scape.getModel()).getNumPlayers() < ((EthnoModel) scape.getModel()).getMaxAgents()) {
                HostCell potentialLoc = (HostCell) ((ScapeArray2DVonNeumann) ((HostCell) this.getHostCell()).getScape()).findRandomUnoccupiedCell();
                EthnoPlayer child = (EthnoPlayer) this.clone();
                ((ScapeVector) getScape()).addAgent(child);
                child.moveTo(potentialLoc);
          
                //mutate strategyToOther?
                if (getRandom().nextFloat() < ((EthnoModel) scape.getModel()).getMutationRate()) {
                    child.strategyToOwn = !(this.strategyToOwn);
                }
                else {
                    child.strategyToOwn = this.strategyToOwn;
                }
                
                //mutate strategyToOther?
                if (getRandom().nextFloat() < ((EthnoModel) scape.getModel()).getMutationRate()) {
                    child.strategyToOther = !(this.strategyToOther);    
                }
                else {
                    child.strategyToOther = this.strategyToOther;
                }
    
                //mutate tag?
                if ((((EthnoModel) scape.getModel()).getOneTagOnly())) {
                    child.tag = this.tag;
                }
                else {
                    if (getRandom().nextFloat() < ((EthnoModel) scape.getModel()).getMutationRate()) {
                        
                        if (((EthnoModel) scape.getModel()).getUseMoreColors()) {
                			int numColors = ((EthnoModel) scape.getModel()).getNumColors();
        
                            child.tag = randomInRange(0,numColors);  //0,3 normal  //0,7 for 8 colors
                            
                            while (child.tag == this.tag) {
                                child.tag = randomInRange(0,numColors);  //0,3 normal  //0,7 for 8 colors
                            }
                            
                        }
                        else {
                            if (this.tag == BLUE) {
                                child.tag = RED;
                            }
                            else {
                                child.tag = BLUE;
                            }                    
                        }
                    }
                    else {
                        child.tag = this.tag;
                    }
                }
                
                ((EthnoCell) child.getHostCell()).updateCell();
                
            }//end if avail
         }//end else random unoccupied site
                
    }//end fission()
    
    
    
    public void play(Agent partner) {
        
        boolean strategy = false;
        boolean partnerStrategy = false;
                   
        //DECIDE WHICH STRATEGY BIT TO USE
        if (this.tag == ((EthnoPlayer) partner).tag) {  //both the same tag, so use "OWN" strategy
                
                strategy = this.strategyToOwn;
        
                //If there's noise, check to see if triggered to switch strategy bit used
                if (((EthnoModel) scape.getModel()).getUseNoise()) {
                    if (randomInRange(0.0,1.0) < ((EthnoModel) scape.getModel()).getNoiseLevel()) {
                        strategy = this.strategyToOther;
                    }
                }    
                
        }
        else { //different tags, use "OTHER" strategy
                
                strategy = this.strategyToOther;
                
                //If there's noise, check to see if triggered to switch strategy bit used
                if (((EthnoModel) scape.getModel()).getUseNoise()) {
                    if (randomInRange(0.0,1.0) < ((EthnoModel) scape.getModel()).getNoiseLevel()) {
                        strategy = this.strategyToOwn;
                    }
                }    
                
        }
                
        //UPDATE BEHAVIOR STATISTICS
        if (this.tag == ((EthnoPlayer) partner).tag) { //they're same type 
            ((EthnoModel) scape.getModel()).incrementPhenotype(MEETOWN);
            
            if (strategy) {
                ((EthnoModel) scape.getModel()).incrementPhenotype(COOPOWN);    
            }
            
        }
        else { //they're different types
            ((EthnoModel) scape.getModel()).incrementPhenotype(MEETOTHER);
            
            if (!(strategy)) {
                ((EthnoModel) scape.getModel()).incrementPhenotype(DEFOTHER);
            }
            else {
                ((EthnoModel) scape.getModel()).incrementPhenotype(COOPOTHER);
            }
                
        }
        
        
        //UPDATE OTRs BASED ON GIVING
        if (strategy == EthnoModel.COOPERATE) {
            this.currOTR = this.currOTR - ((EthnoModel) scape.getModel()).getCostOfGiving();
            ((EthnoPlayer) partner).currOTR = ((EthnoPlayer) partner).currOTR + ((EthnoModel) scape.getModel()).getGainOfReceiving();
        }
        else {
        	
        	if (((EthnoModel) scape.getModel()).getUseCombat()) {
        		
        		this.currOTR = this.currOTR + ((EthnoModel) scape.getModel()).getCostOfGiving();
        		((EthnoPlayer) partner).currOTR = ((EthnoPlayer) partner).currOTR - ((EthnoModel) scape.getModel()).getGainOfReceiving();	
        		
        	}	
        	
        }
    }
    
    
    
	public Color getColor() {
	    
	    //usual color scheme
		if (tag == BLUE) {
			return Color.blue;
		}
		if (tag == RED) {
		    return Color.red;
		}
		if (tag == GREEN) {
		    return Color.green;
		}
		if (tag == YELLOW) {
		    return Color.yellow;
		}
		else {
			return Color.black; //default, if appears means error
		}
    }
    
    
    //GETTERS AND SETTERS
    
    public int getTag() {
        return (this.tag);
    }
    
    public int getStrategyToOwn() {
        if (this.strategyToOwn) {
            return (0); //this means cooperate
        }
        else {
            return (1); //this means defect
        }   
    }
    public void setStrategyToOwn(int _strategy) {
        
    }
     
    public int getStrategyToOther() {
        if (this.strategyToOther) {
            return (0); //this means cooperate
        }
        else {
            return (1); //this means defect
        }   
    }
    public void setStrategyToOther(int _strategy) {
 
    }
    public boolean getStrategyOwn() {
        return (this.strategyToOwn);
    }
    public boolean getStrategyOther() {
        return (this.strategyToOther);
    }
    
}